var _this = this;

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

import { text } from 'd3-request';
import { bisector } from 'd3-array';
import { tsvParseRows } from 'd3-dsv';
import { scaleLinear, scaleLog } from 'd3-scale';
import { expose, Transfer } from 'threads/worker';
import { TabixIndexedFile } from '@gmod/tabix';
import VCF from '@gmod/vcf';
import { RemoteFile } from 'generic-filehandle';
import LRU from 'lru-cache';
import slugid from 'slugid';
import { PILEUP_COLOR_IXS } from './vcf-utils';

function currTime() {
  var d = new Date();
  return d.getTime();
}
/////////////////////////////////////////////////
/// ChromInfo
/////////////////////////////////////////////////

var chromInfoBisector = bisector(function (d) {
  return d.pos;
}).left;

var chrToAbs = function chrToAbs(chrom, chromPos, chromInfo) {
  return chromInfo.chrPositions[chrom].pos + chromPos;
};

var absToChr = function absToChr(absPosition, chromInfo) {
  if (!chromInfo || !chromInfo.cumPositions || !chromInfo.cumPositions.length) {
    return null;
  }

  var insertPoint = chromInfoBisector(chromInfo.cumPositions, absPosition);
  var lastChr = chromInfo.cumPositions[chromInfo.cumPositions.length - 1].chr;
  var lastLength = chromInfo.chromLengths[lastChr];

  insertPoint -= insertPoint > 0 && 1;

  var chrPosition = Math.floor(absPosition - chromInfo.cumPositions[insertPoint].pos);
  var offset = 0;

  if (chrPosition < 0) {
    // before the start of the genome
    offset = chrPosition - 1;
    chrPosition = 1;
  }

  if (insertPoint === chromInfo.cumPositions.length - 1 && chrPosition > lastLength) {
    // beyond the last chromosome
    offset = chrPosition - lastLength;
    chrPosition = lastLength;
  }

  return [chromInfo.cumPositions[insertPoint].chr, chrPosition, offset, insertPoint];
};

function parseChromsizesRows(data) {
  var cumValues = [];
  var chromLengths = {};
  var chrPositions = {};

  var totalLength = 0;

  for (var i = 0; i < data.length; i++) {
    var length = Number(data[i][1]);
    totalLength += length;

    var newValue = {
      id: i,
      chr: data[i][0],
      pos: totalLength - length
    };

    cumValues.push(newValue);
    chrPositions[newValue.chr] = newValue;
    chromLengths[data[i][0]] = length;
  }

  return {
    cumPositions: cumValues,
    chrPositions: chrPositions,
    totalLength: totalLength,
    chromLengths: chromLengths
  };
}

function ChromosomeInfo(filepath, success) {
  var ret = {};

  ret.absToChr = function (absPos) {
    return ret.chrPositions ? absToChr(absPos, ret) : null;
  };

  ret.chrToAbs = function () {
    var _ref = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : [],
        _ref2 = _slicedToArray(_ref, 2),
        chrName = _ref2[0],
        chrPos = _ref2[1];

    return ret.chrPositions ? chrToAbs(chrName, chrPos, ret) : null;
  };

  return text(filepath, function (error, chrInfoText) {
    if (error) {
      // console.warn('Chromosome info not found at:', filepath);
      if (success) success(null);
    } else {
      var data = tsvParseRows(chrInfoText);
      var chromInfo = parseChromsizesRows(data);

      Object.keys(chromInfo).forEach(function (key) {
        ret[key] = chromInfo[key];
      });
      if (success) success(ret);
    }
  });
}

/////////////////////////////////////////////////////
/// End Chrominfo
/////////////////////////////////////////////////////

var extractColumnFromVcfInfo = function extractColumnFromVcfInfo(info, index) {
  var col = {};
  Object.keys(info).forEach(function (key) {
    col[key] = info[key][index];
  });
  return col;
};

var vcfRecordToJson = function vcfRecordToJson(vcfRecord, chrName, chrOffset) {
  var segments = [];
  var info = vcfRecord['INFO'];

  // VCF records can have multiple ALT. We create a segment for each of them
  vcfRecord['ALT'].forEach(function (alt, index) {
    var segment = {
      id: slugid.nice(),
      alt: alt,
      ref: vcfRecord.REF,
      from: vcfRecord.POS + chrOffset,
      to: chrOffset,
      chrName: chrName,
      chrOffset: chrOffset,
      alleleCount: info.AC[index],
      alleleFrequency: info.AF[index],
      alleleNumber: info.AN[index],
      info: extractColumnFromVcfInfo(info, index),
      row: null,
      type: 'variant',
      category: 'SNV'
    };

    // CNVs
    if (alt.includes(">")) {
      //console.log(vcfRecord);
      segment['to'] += info.END[index];
      segment['category'] = 'CNV';
      if (info.SVTYPE[index] === "INS") {
        segment['type'] = 'insertion';
      } else if (info.SVTYPE[index] === "DEL") {
        segment['type'] = 'deletion';
      } else if (info.SVTYPE[index] === "DUP") {
        segment['type'] = 'duplication';
      } else if (info.SVTYPE[index] === "INV") {
        segment['type'] = 'inversion';
      }
      //console.log(segment);
    }
    // SNVs
    else {
        segment['to'] += vcfRecord.POS + vcfRecord.REF.length;
        if (segment['alt'].length > segment['ref'].length) {
          segment['type'] = 'insertion';
        } else if (segment['alt'].length < segment['ref'].length) {
          segment['type'] = 'deletion';
        }
      }

    segments.push(segment);
  });

  return segments;
};

// promises indexed by urls
var vcfFiles = {};
var vcfHeaders = {};
var tbiVCFParsers = {};

var MAX_TILES = 20;

// promises indexed by url
var chromSizes = {};
var chromInfos = {};
var tileValues = new LRU({ max: MAX_TILES });
var tilesetInfos = {};

// indexed by uuid
var dataConfs = {};

var init = function init(uid, vcfUrl, tbiUrl, chromSizesUrl, maxTileWidth) {
  if (!vcfFiles[vcfUrl]) {
    vcfFiles[vcfUrl] = new TabixIndexedFile({
      filehandle: new RemoteFile(vcfUrl),
      tbiFilehandle: new RemoteFile(tbiUrl)
    });

    vcfHeaders[vcfUrl] = vcfFiles[vcfUrl].getHeader();
    // vcfFiles[vcfUrl].getHeader().then(headerText => {
    //   vcfHeaders[vcfUrl] = headerText;
    //   tbiVCFParsers[vcfUrl] = new VCF({ header: headerText });

    // });
  }

  if (chromSizesUrl) {
    chromSizes[chromSizesUrl] = chromSizes[chromSizesUrl] || new Promise(function (resolve) {
      ChromosomeInfo(chromSizesUrl, resolve);
    });
  }

  dataConfs[uid] = {
    vcfUrl: vcfUrl,
    chromSizesUrl: chromSizesUrl,
    maxTileWidth: maxTileWidth
  };
};

var tilesetInfo = function tilesetInfo(uid) {
  var _dataConfs$uid = dataConfs[uid],
      chromSizesUrl = _dataConfs$uid.chromSizesUrl,
      vcfUrl = _dataConfs$uid.vcfUrl;

  var promises = [vcfHeaders[vcfUrl], chromSizes[chromSizesUrl]];

  return Promise.all(promises).then(function (values) {
    if (!tbiVCFParsers[vcfUrl]) {
      tbiVCFParsers[vcfUrl] = new VCF({ header: values[0] });
    }

    var TILE_SIZE = 1024;
    var chromInfo = values[1];
    chromInfos[chromSizesUrl] = chromInfo;

    var retVal = {
      tile_size: TILE_SIZE,
      bins_per_dimension: TILE_SIZE,
      max_zoom: Math.ceil(Math.log(chromInfo.totalLength / TILE_SIZE) / Math.log(2)),
      max_width: chromInfo.totalLength,
      min_pos: [0],
      max_pos: [chromInfo.totalLength]
    };

    tilesetInfos[uid] = retVal;
    return retVal;
  });
};

var tile = function () {
  var _ref3 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee(uid, z, x) {
    var _dataConfs$uid2, vcfUrl, chromSizesUrl, maxTileWidth, vcfFile;

    return regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            _dataConfs$uid2 = dataConfs[uid], vcfUrl = _dataConfs$uid2.vcfUrl, chromSizesUrl = _dataConfs$uid2.chromSizesUrl, maxTileWidth = _dataConfs$uid2.maxTileWidth;
            vcfFile = vcfFiles[vcfUrl];
            return _context.abrupt('return', tilesetInfo(uid).then(function (tsInfo) {
              var tileWidth = +tsInfo.max_width / Math.pow(2, +z);
              var recordPromises = [];

              if (tileWidth > maxTileWidth) {
                return new Promise(function (resolve) {
                  return resolve([]);
                });
              }

              // get the bounds of the tile
              var minX = tsInfo.min_pos[0] + x * tileWidth;
              var maxX = tsInfo.min_pos[0] + (x + 1) * tileWidth;

              var chromInfo = chromInfos[chromSizesUrl];

              var chromLengths = chromInfo.chromLengths,
                  cumPositions = chromInfo.cumPositions;


              var variants = [];

              var _loop = function _loop(i) {
                var chromName = cumPositions[i].chr;
                var chromStart = cumPositions[i].pos;

                var chromEnd = cumPositions[i].pos + chromLengths[chromName];
                tileValues.set(uid + '.' + z + '.' + x, []);

                if (chromStart <= minX && minX < chromEnd) {
                  // start of the visible region is within this chromosome

                  if (maxX > chromEnd) {
                    // the visible region extends beyond the end of this chromosome
                    // fetch from the start until the end of the chromosome
                    var startPos = minX - chromStart;
                    var endPos = chromEnd - chromStart;

                    recordPromises.push(vcfFile.getLines(chromName, startPos, endPos, function (line) {
                      var vcfRecord = tbiVCFParsers[vcfUrl].parseLine(line);
                      var vcfJson = vcfRecordToJson(vcfRecord, chromName, cumPositions[i].pos);
                      vcfJson.forEach(function (variant) {
                        return variants.push(variant);
                      });
                    }));
                    minX = chromEnd;
                  } else {
                    var _endPos = Math.ceil(maxX - chromStart);
                    var _startPos = Math.floor(minX - chromStart);

                    recordPromises.push(vcfFile.getLines(chromName, _startPos, _endPos, function (line) {
                      var vcfRecord = tbiVCFParsers[vcfUrl].parseLine(line);
                      var vcfJson = vcfRecordToJson(vcfRecord, chromName, cumPositions[i].pos);
                      vcfJson.forEach(function (variant) {
                        return variants.push(variant);
                      });
                    }));

                    // end the loop because we've retrieved the last chromosome
                    return 'break';
                  }
                }
              };

              for (var i = 0; i < cumPositions.length; i++) {
                var _ret = _loop(i);

                if (_ret === 'break') break;
              }

              // flatten the array of promises so that it looks like we're
              // getting one long list of value
              return Promise.all(recordPromises).then(function () {
                //console.log(variants);
                tileValues.set(uid + '.' + z + '.' + x, variants);
                return variants;
              });
            }));

          case 3:
          case 'end':
            return _context.stop();
        }
      }
    }, _callee, _this);
  }));

  return function tile(_x2, _x3, _x4) {
    return _ref3.apply(this, arguments);
  };
}();

var fetchTilesDebounced = function () {
  var _ref4 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee2(uid, tileIds) {
    var tiles, validTileIds, tilePromises, _iteratorNormalCompletion, _didIteratorError, _iteratorError, _iterator, _step, tileId, parts, z, x;

    return regeneratorRuntime.wrap(function _callee2$(_context2) {
      while (1) {
        switch (_context2.prev = _context2.next) {
          case 0:
            tiles = {};
            validTileIds = [];
            tilePromises = [];
            _iteratorNormalCompletion = true;
            _didIteratorError = false;
            _iteratorError = undefined;
            _context2.prev = 6;
            _iterator = tileIds[Symbol.iterator]();

          case 8:
            if (_iteratorNormalCompletion = (_step = _iterator.next()).done) {
              _context2.next = 21;
              break;
            }

            tileId = _step.value;
            parts = tileId.split('.');
            z = parseInt(parts[0], 10);
            x = parseInt(parts[1], 10);

            if (!(Number.isNaN(x) || Number.isNaN(z))) {
              _context2.next = 16;
              break;
            }

            console.warn('Invalid tile zoom or position:', z, x);
            return _context2.abrupt('continue', 18);

          case 16:

            validTileIds.push(tileId);
            tilePromises.push(tile(uid, z, x));

          case 18:
            _iteratorNormalCompletion = true;
            _context2.next = 8;
            break;

          case 21:
            _context2.next = 27;
            break;

          case 23:
            _context2.prev = 23;
            _context2.t0 = _context2['catch'](6);
            _didIteratorError = true;
            _iteratorError = _context2.t0;

          case 27:
            _context2.prev = 27;
            _context2.prev = 28;

            if (!_iteratorNormalCompletion && _iterator.return) {
              _iterator.return();
            }

          case 30:
            _context2.prev = 30;

            if (!_didIteratorError) {
              _context2.next = 33;
              break;
            }

            throw _iteratorError;

          case 33:
            return _context2.finish(30);

          case 34:
            return _context2.finish(27);

          case 35:
            return _context2.abrupt('return', Promise.all(tilePromises).then(function (values) {
              for (var i = 0; i < values.length; i++) {
                var validTileId = validTileIds[i];
                tiles[validTileId] = values[i];
                tiles[validTileId].tilePositionId = validTileId;
              }

              return tiles;
            }));

          case 36:
          case 'end':
            return _context2.stop();
        }
      }
    }, _callee2, _this, [[6, 23, 27, 35], [28,, 30, 34]]);
  }));

  return function fetchTilesDebounced(_x5, _x6) {
    return _ref4.apply(this, arguments);
  };
}();

///////////////////////////////////////////////////
/// Render Functions
///////////////////////////////////////////////////

var STARTING_POSITIONS_ARRAY_LENGTH = Math.pow(2, 20);
var STARTING_COLORS_ARRAY_LENGTH = Math.pow(2, 21);
var STARTING_INDEXES_LENGTH = Math.pow(2, 21);

var allPositionsLength = STARTING_POSITIONS_ARRAY_LENGTH;
var allColorsLength = STARTING_COLORS_ARRAY_LENGTH;
var allIndexesLength = STARTING_INDEXES_LENGTH;

var allPositions = new Float32Array(allPositionsLength);
var allColors = new Float32Array(allColorsLength);
var allIndexes = new Int32Array(allIndexesLength);

var renderSegments = function renderSegments(uid, tileIds, domain, scaleRange, trackOptions, labelPositions) {
  //const t1 = currTime();
  var allSegments = {};

  var _iteratorNormalCompletion2 = true;
  var _didIteratorError2 = false;
  var _iteratorError2 = undefined;

  try {
    for (var _iterator2 = tileIds[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
      var tileId = _step2.value;

      var tileValue = tileValues.get(uid + '.' + tileId);

      if (!tileValue) continue;

      if (tileValue.error) {
        throw new Error(tileValue.error);
      }

      var _iteratorNormalCompletion3 = true;
      var _didIteratorError3 = false;
      var _iteratorError3 = undefined;

      try {
        for (var _iterator3 = tileValue[Symbol.iterator](), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
          var segment = _step3.value;

          allSegments[segment.id] = segment;
        }
      } catch (err) {
        _didIteratorError3 = true;
        _iteratorError3 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion3 && _iterator3.return) {
            _iterator3.return();
          }
        } finally {
          if (_didIteratorError3) {
            throw _iteratorError3;
          }
        }
      }
    }
  } catch (err) {
    _didIteratorError2 = true;
    _iteratorError2 = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion2 && _iterator2.return) {
        _iterator2.return();
      }
    } finally {
      if (_didIteratorError2) {
        throw _iteratorError2;
      }
    }
  }

  var segmentList = Object.values(allSegments);

  var _ref5 = [Number.MAX_VALUE, -Number.MAX_VALUE],
      minPos = _ref5[0],
      maxPos = _ref5[1];


  for (var i = 0; i < segmentList.length; i++) {
    if (segmentList[i].from < minPos) {
      minPos = segmentList[i].from;
    }

    if (segmentList[i].to > maxPos) {
      maxPos = segmentList[i].to;
    }
  }

  var currPosition = 0;
  var currColor = 0;
  var currIdx = 0;

  var addPosition = function addPosition(x1, y1) {
    if (currPosition > allPositionsLength - 2) {
      allPositionsLength *= 2;
      var prevAllPositions = allPositions;

      allPositions = new Float32Array(allPositionsLength);
      allPositions.set(prevAllPositions);
    }
    allPositions[currPosition++] = x1;
    allPositions[currPosition++] = y1;

    return currPosition / 2 - 1;
  };

  var addColor = function addColor(colorIdx, n) {
    if (currColor >= allColorsLength - n) {
      allColorsLength *= 2;
      var prevAllColors = allColors;

      allColors = new Float32Array(allColorsLength);
      allColors.set(prevAllColors);
    }

    for (var k = 0; k < n; k++) {
      allColors[currColor++] = colorIdx;
    }
  };

  var addTriangleIxs = function addTriangleIxs(ix1, ix2, ix3) {
    if (currIdx >= allIndexesLength - 3) {
      allIndexesLength *= 2;
      var prevAllIndexes = allIndexes;

      allIndexes = new Int32Array(allIndexesLength);
      allIndexes.set(prevAllIndexes);
    }

    allIndexes[currIdx++] = ix1;
    allIndexes[currIdx++] = ix2;
    allIndexes[currIdx++] = ix3;
  };

  var addRect = function addRect(x, y, width, height, colorIdx) {
    var xLeft = x;
    var xRight = xLeft + width;
    var yTop = y;
    var yBottom = y + height;

    var ulIx = addPosition(xLeft, yTop);
    var urIx = addPosition(xRight, yTop);
    var llIx = addPosition(xLeft, yBottom);
    var lrIx = addPosition(xRight, yBottom);
    addColor(colorIdx, 4);

    addTriangleIxs(ulIx, urIx, llIx);
    addTriangleIxs(llIx, lrIx, urIx);
  };

  var xScale = scaleLinear().domain(domain).range(scaleRange);

  var pos0 = labelPositions[0];
  var pos10m5 = labelPositions[1];
  var pos10m0 = labelPositions[labelPositions.length - 1];

  var logYScale10m5 = scaleLog().domain([1e-5, 1]).range([pos10m5, pos10m0]);
  var logYScale10m9 = scaleLog().domain([1e-8, 1e-5]).range([pos0, pos10m5]);

  var xLeft = void 0;
  var xRight = void 0;
  var yTop = void 0;

  // Needed to check for duplicates
  segmentList.sort(function (a, b) {
    return a.from - b.from;
  });
  var lastSegment = null;

  segmentList.forEach(function (segment, j) {
    // Ignore duplicates - can happen when variants span more than one tile
    if (lastSegment && segment.from === lastSegment.from && segment.ref === lastSegment.ref && segment.alt === lastSegment.alt) {
      return;
    }
    lastSegment = segment;

    var from = xScale(segment.from);
    var to = xScale(segment.to);

    if (segment.alleleFrequency >= 1e-5) {
      yTop = logYScale10m5(segment.alleleFrequency) - trackOptions.variantHeight / 2;
    } else if (segment.alleleFrequency >= 1e-8) {
      yTop = logYScale10m9(segment.alleleFrequency) - trackOptions.variantHeight / 2;
    } else {
      yTop = 0;
    }
    // Shift everything by one, since the graphics starts at 1
    yTop += 1;

    var width = to - from;
    // This is needed because a constant padding would be too large, if the
    // initial rendering is happing zoomed out
    var padding = Math.min(0.5, 0.01 * width);

    xLeft = from + padding;
    xRight = to - padding;

    var colorToUse = PILEUP_COLOR_IXS.VARIANT;
    if (segment.type === 'deletion') {
      colorToUse = PILEUP_COLOR_IXS.DELETION;
    } else if (segment.type === 'insertion') {
      colorToUse = PILEUP_COLOR_IXS.INSERTION;
    } else if (segment.type === 'duplication') {
      colorToUse = PILEUP_COLOR_IXS.DUPLICATION;
    } else if (segment.type === 'inversion') {
      colorToUse = PILEUP_COLOR_IXS.INVERSION;
    }
    segment['yTop'] = yTop;
    addRect(xLeft, yTop, xRight - xLeft, trackOptions.variantHeight, colorToUse);
  });

  var positionsBuffer = allPositions.slice(0, currPosition).buffer;
  var colorsBuffer = allColors.slice(0, currColor).buffer;
  var ixBuffer = allIndexes.slice(0, currIdx).buffer;

  var objData = {
    variants: segmentList,
    positionsBuffer: positionsBuffer,
    colorsBuffer: colorsBuffer,
    ixBuffer: ixBuffer,
    xScaleDomain: domain,
    xScaleRange: scaleRange
  };

  //const t2 = currTime();
  //console.log('renderSegments time:', t2 - t1, 'ms');

  return Transfer(objData, [objData.positionsBuffer, colorsBuffer, ixBuffer]);
};

var tileFunctions = {
  init: init,
  tilesetInfo: tilesetInfo,
  fetchTilesDebounced: fetchTilesDebounced,
  tile: tile,
  renderSegments: renderSegments
};

expose(tileFunctions);